@@ -6,9 +6,9 @@ Created on Fri Nov 20 20:22:44 2020
 """
 
 # Python通用工具库，春秋迭代，yondersky@126.com，2020-11-20
-# 更新日期：2022-02-23
+# 更新日期：2022-04-14
 
-import collections, functools, itertools, math, numbers, orderedset, time
+import collections, functools, itertools, math, numbers, time
 import datetime as dt
 import numpy as np
 import pandas as pd
@@ -91,8 +91,8 @@ def emap(func):
                 return result
             elif returnType is np.ndarray:
                 return np.array(list(result))
-            elif isinstance(iterables,Series) and returnType==Series:
-                return Series(result,iterables.index)
+            elif hasattr(returnType,'values') and hasattr(iterables,'values'):
+                return returnType(list(result),iterables.index)
             else:
                 return returnType(result)
         else:
@@ -120,8 +120,8 @@ def temap(func, returnType = None):
                 return result
             elif returnType is np.ndarray:
                 return np.array(list(result))
-            elif isinstance(iterables,Series) and returnType==Series:
-                return Series(result,iterables.index)
+            elif hasattr(returnType,'values') and hasattr(iterables,'values'):
+                return returnType(result,iterables.index)
             else:
                 return returnType(result)
         else:
@@ -161,8 +161,8 @@ def sapply(func, returnType = None):
                 return result
             elif rttype is np.ndarray:
                 return np.array(result)
-            elif isinstance(iterables,Series) and returnType==Series:
-                return Series(result,iterables.index)
+            elif hasattr(rttype,'values') and hasattr(iterables,'values'):
+                return rttype(result,iterables.index)
             else:
                 return rttype(result)
         else:
@@ -191,8 +191,8 @@ def tsapply(func, returnType = None):
                 return result
             elif rttype is np.ndarray:
                 return np.array(result)
-            elif isinstance(iterables,Series) and returnType==Series:
-                return Series(result,iterables.index)
+            elif hasattr(rttype,'values') and hasattr(iterables,'values'):
+                return rttype(result,iterables.index)
             else:
                 return rttype(result)
         else:
@@ -202,6 +202,7 @@ def tsapply(func, returnType = None):
 
 # 2.1.3 多参数函数（向量化全部无名称参数）
 
+# 普通函数版本
 # 2021-12-18
 def fapply(func, returnType = None):
     '''
@@ -238,8 +239,8 @@ def fapply(func, returnType = None):
                 return result
             elif rttype is np.ndarray:
                 return np.array(result)
-            elif rttype==Series and isinstance(iterables[0],Series):
-                return Series(result,iterables[0].index)
+            elif hasattr(rttype,'values') and hasattr(iterables[0],'values'):
+                return rttype(result,iterables[0].index)
             else:
                 return rttype(result)
         else:
@@ -247,6 +248,37 @@ def fapply(func, returnType = None):
     
     return applied
 
+# 类成员函数版本
+# 2022-04-06
+def tfapply(func, returnType = None):
+    '''类成员函数版本的fapply装饰器'''
+    @functools.wraps(func)
+    
+    # 2022-04-06
+    def applied(self, *iterables, **kwargs):
+        iterable = True
+        for i in iterables:
+            if IsSingleType(i):
+                iterable = False
+                break
+        if iterable:
+            result = [func(self,*zipi,**kwargs) for zipi in zip(*iterables)]
+            rttype = returnType
+            if rttype is None:
+                rttype = type(iterables[0])
+            if rttype is range or rttype is set:
+                rttype = list
+            if rttype is list:
+                return result
+            elif rttype is np.ndarray:
+                return np.array(result)
+            elif hasattr(rttype,'values') and hasattr(iterables[0],'values'):
+                return rttype(result,iterable[0].index)
+            else:
+                return rttype(result)
+        else:
+            return func(self,*iterables,**kwargs)
+
 # 3. 格式转换
 
 # 3.1 字符串相关
@@ -374,6 +406,36 @@ def CurrentTimeStr():
     '''当前时间戳'''
     return time.strftime('%Y-%m-%d %H:%M:%S')
 
+# 3.1.6 可迭代对象转字符串
+
+# 2022-04-14
+def ListToStr(iterables, sep = ',', unique = False, NAIgnored = False):
+    '''
+    可迭代对象转字符串。
+
+    示例：
+    >>> ListToStr([1,2,3],';')
+    '1;2;3'
+    >>> ListToStr(Series([1,3,2,3,2,1]),unique=True)
+    '1,3,2'
+    >>> ListToStr('abc')
+    'abc'
+    >>> ListToStr(['a','b',NA,'c'])
+    'a,b,NA,c'
+    >>> ListToStr(['a','b',NA,'c'],NAIgnored=True)
+    'a,b,c'
+    >>> ListToStr({'a':'aa','b':'bb'})
+    'aa,bb'
+    '''
+    values = Series(iterables)
+    if unique:
+        values = values.drop_duplicates()
+    if NAIgnored:
+        values = values.dropna()
+    else:
+        values = values.replace(np.nan,'NA')
+    return sep.join(values.astype(str))
+
 # 3.2 数字相关
 
 # 3.2.1 舍入
@@ -792,6 +854,22 @@ def EndOfTheWeek(date, weekStart = 1):
     end = date+dt.timedelta((-date.weekday()+weekStart-2)%7)
     return DateToInt(end) if dateInt else end
 
+# 3.6.3.5 日期间隔
+
+# 2022-04-02
+def DateDiff(startDate, endDate):
+    '''
+    日期间隔。
+    【示例】
+    >>> DateDiff(20220101,20220331)
+    89
+    '''
+    if IsInteger(startDate):
+        startDate = IntToDate(startDate)
+    if IsInteger(endDate):
+        endDate = IntToDate(endDate)
+    return (endDate-startDate).days
+
 # 4. 类型判断
 
 # 4.1 变量类型判断
@@ -1170,11 +1248,11 @@ def rank(x, ascending = True, na_option = 'keep', method = 'average'):
 # 5.2.2 百分比排名
 
 # 2021-10-14
-def PercentRank(x, ascending = True, na_option = 'keep', na_default = 0.5,
-    method = 'average', bound_method = 'mid'):
+def PercentRank(x, ascending = True, na_option = 'keep',  method = 'average',
+     na_default = 0.5, bound_method = 'mid'):
     '''
     百分比排名。
-    注：
+    【注】
     1. ascending、na_option、method参数请参见rank函数对应同名参数的帮助；
     2. 仅当na_option参数取值为keep时，na_default参数可用，将排名后的NA值以
        na_default替换；
@@ -1182,55 +1260,26 @@ def PercentRank(x, ascending = True, na_option = 'keep', na_default = 0.5,
       exc - 极值在边界
       inc - 极值在边界内一区间
       mid - 极值在边界内中点处
-    4. x必需为Series或DataFrame。
-
-    示例：
-    >>> s1 = Series([1,2,3,NA,5])
-    >>> PercentRank(s1)
-    0    0.125
-    1    0.375
-    2    0.625
-    3    0.500
-    4    0.875
-    dtype: float64
-    >>> PercentRank(s1,bound_method='exc')
-    0    0.000000
-    1    0.333333
-    2    0.666667
-    3    0.500000
-    4    1.000000
-    dtype: float64
-    >>> PercentRank(s1,bound_method='inc')
-    0    0.2
-    1    0.4
-    2    0.6
-    3    0.5
-    4    0.8
-    dtype: float64
+    【示例】
+    >>> PercentRank([1,2,3,NA,5])
+    array([0.125, 0.375, 0.625, 0.5  , 0.875])
+    >>> PercentRank([1,2,3,NA,5],bound_method='exc')
+    array([0.        , 0.33333333, 0.66666667, 0.5       , 1.        ])
+    >>> PercentRank([1,2,3,NA,5],bound_method='inc')
+    array([0.2, 0.4, 0.6, 0.5, 0.8])
     '''
-    rt = x.rank(method=method,na_option=na_option,ascending=ascending)
-    if IsDataFrame(rt):
-        for c in rt.columns:
-            clen = len(np.where(~IsNA(rt[c]))[0]) if na_option=='keep' \
-                else rt.shape[0]
-            if clen>1:
-                if bound_method=='exc':
-                    rt[c] = (rt[c]-1)/(clen-1)
-                elif bound_method=='inc':
-                    rt[c] /= (clen+1)
-                else:
-                    rt[c] = (rt[c]-0.5)/clen
-    else:
-        clen = len(np.where(~np.isnan(rt))[0]) if na_option=='keep' else len(rt)
-        if clen>1:
-            if bound_method=='exc':
-                rt = (rt-1)/(clen-1)
-            elif bound_method=='inc':
-                rt /= (clen+1)
-            else:
-                rt = (rt-0.5)/clen
+    rt = rank(x,ascending,na_option,method)
+    napos = which(np.isnan(rt))
+    clen = len(rt)-len(napos) if na_option=='keep' else len(rt)
+    if clen>1:
+        if bound_method=='exc':
+            rt = (rt-1)/(clen-1)
+        elif bound_method=='inc':
+            rt/= (clen+1)
+        else:
+            rt = (rt-0.5)/clen
     if na_option=='keep' and not np.isnan(na_default):
-        rt.fillna(na_default,inplace=True)
+        rt[napos] = na_default
     return rt
 
 # 5.2.3 下标排序
@@ -1425,16 +1474,46 @@ def posdiff(size, pos):
 # 5.3.2 集合取补
 
 # 2022-02-03
-def setdiff(a, b):
+def setdiff(a, b, ordered = True):
     '''
-    集合取补。
+    集合取补（保序）。
     【示例】
-    >>> setdiff(list('abcde'),list('ceg'))
-    ['a', 'b', 'd']
+    >>> setdiff([5,3,2,4,1],[3,5,2])
+    [4, 1]
+    >>> setdiff([5,3,2,4,1],[3,5,2],False)
+    [1, 4]
     '''
-    a = orderedset.OrderedSet(a)
-    b = set(b)
-    return [i for i in a if not i in b]
+    if ordered:
+        a = collections.OrderedDict.fromkeys(a)
+        b = set(b)
+        return [i for i in a if not i in b]
+    else:
+        return list(np.setdiff1d(a,b))
+
+# 5.4 增长率计算
+
+# 2022-03-11
+def GrowthRate(startValue, endValue):
+    '''
+    财务增长率。
+    【公式】（当期值-上期值）/abs（上期值）
+    【示例】
+    >>> GrowthRate([2,-1],[-1,2])
+    array([-1.5,  3. ])
+    '''
+    if hasattr(startValue,'_ValidData'):
+        startValue = startValue._ValidData
+    elif hasattr(startValue,'values'):
+        startValue = startValue.values
+    elif not IsArray(startValue):
+        startValue = np.array(startValue,np.float64)
+    if hasattr(endValue,'_ValidData'):
+        endValue = endValue._ValidData
+    elif hasattr(endValue,'values'):
+        endValue = endValue.values
+    elif not IsArray(endValue):
+        endValue = np.array(endValue,np.float64)
+    return (endValue-startValue)/np.abs(startValue)
 
 # 6. 逻辑扩展
 
